# Copyright (C) 2016  Pachol, Vojtěch <pacholick@gmail.com>
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

from colorama import Style
from itertools import zip_longest


class PrettyNumber:
    """Class for making pretty console output of a number."""

    def _swap_color_functions(self):
        p = self.get_color
        self.get_color = self.get_reset
        self.get_reset = p

    def _reset_color_functions(self):
        self.get_color = self.GET_COLOR
        self.get_reset = self.GET_RESET

    def get_color(self):
        """Get highlighting string and swap with reset."""
        self._swap_color_functions()
        return Style.BRIGHT

    def get_reset(self):
        """Get reset string and swap with color."""
        self._swap_color_functions()
        return Style.RESET_ALL

    def split(self):
        """Split the number into its parts."""
        # s = str(self.NUMBER)
        s = '{:.12g}'.format(self.NUMBER)
        self.int_part, self.dot, self.frac_part = s.partition(".")
        _, self.minus, self.int_part = self.int_part.rpartition("-")
        self.frac_part, self.e, self.exponent = self.frac_part.partition("e")
        _, self.exp_plusminus, self.exponent = self.exponent.rpartition("-")
        if self.exp_plusminus == '':
            _, self.exp_plusminus, self.exponent = self.exponent.rpartition(
                "+")

    def color_thousands(self):
        """Color thousands in integer and fractional parts of the number."""
        lst = []
        for i in map(lambda s: ''.join(reversed(s)),
                     zip_longest(*[reversed(self.int_part)]*3, fillvalue='')):
            lst.append(i)
            lst.append(self.get_reset())
        lst.reverse()
        self.int_part = ''.join(lst)

        lst = []
        self._reset_color_functions()
        for i in map(''.join,
                     zip_longest(*[iter(self.frac_part)]*3, fillvalue='')):
            lst.append(self.get_reset())
            lst.append(i)
        lst.append(self.get_reset())
        self.frac_part = ''.join(lst)

    def __init__(self, number):
        """
        :param number: The number to print
        """
        self.GET_COLOR = self.get_color
        self.GET_RESET = self.get_reset

        self.NUMBER = number
        self.split()
        self.color_thousands()

    def __str__(self):
        return ''.join((self.minus, self.int_part, self.dot, self.frac_part,
                        self.e, self.exp_plusminus, self.exponent))



# Copyright (C) 2016  Pachol, Vojtěch <pacholick@gmail.com>
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

import re as _re
import math as _math


_SYMBOLS = {
    'da': "10",     # deca
    'h': "100",     # hecto
    'k': "1000",    # kilo
    'M': "1000000",     # mega
    'G': "1000000000",  # giga
    'T': "1000000000000",   # tera
    'P': "1000000000000000",    # peta
    'E': "1000000000000000000",     # exa
    'Z': "1000000000000000000000",  # zetta
    'Y': "1000000000000000000000000",   # yotta
    'R': "1000000000000000000000000000",    # ronna
    'Q': "1000000000000000000000000000000",     # quecca

    'd': "0.1",     # deci
    'c': "0.01",    # centi
    'm': "0.001",   # mili
    'μ': "0.000001",    # micro (greek mu)
    'µ': "0.000001",    # micro (micro sign)
    'u': "0.000001",    # micro (‘u’ is also allowed)
    'n': "0.000000001",     # nano
    'p': "0.000000000001",  # pico
    'f': "0.000000000000001",   # femto
    'a': "0.000000000000000001",    # atto
    'z': "0.000000000000000000001",     # zepto
    'y': "0.000000000000000000000001",  # yocto
    'r': "0.000000000000000000000000001",   # ronto
    'q': "0.000000000000000000000000000001",    # quecto

    '°': "(π/180)",
}


def replace_symbols(input_):
    """replace 123k with 1000*123"""
    for symbol, value in _SYMBOLS.items():
        input_ = _re.sub(
            r'(\d+){}'.format(symbol),
            r'\1*{}'.format(value),
            input_
        )
    return input_


def from_fahrenheit(fahrenheit):
    """Convert Fahrenheits to Centigrade."""
    return (fahrenheit - 32)*5/9


C = from_fahrenheit


def to_fahrenheit(c):
    """Convert Centigrade to Fahrenheits."""
    return c*9/5 + 32


def prefix(x, dimension=1):
    """Give the number an appropriate SI prefix.

    :param x:   Too big or too small number.
    :returns:   String containing a number between 1 and 1000 and SI prefix.
    """
    if x == 0:
        return "0  "

    length = _math.floor(_math.log10(abs(x)))
    if abs(length) > 30:
        length = int(_math.copysign(30, length))

    div, mod = divmod(length, 3*dimension)
    return "{:.3g} {}".format(
        x*10**(-length + mod),
        " kMGTPEZYRQqryzafpnµm"[div])


# maths
π = _math.pi
φ = (1 + _math.sqrt(5))/2


# time
days_in_year = 365.242189       # tropical year
# days_in_year = 365.256363004    # sideral year
seconds_in_day = 86400
year = yr = seconds_in_year = days_in_year*seconds_in_day


# physics
c = 299792458
"""Speed of light in vacuum (exact) [ m s⁻¹ ]"""
G = 6.6740831e-11
"""Gravitational constant [ m³ kg⁻¹ s⁻² ]"""
h = 6.62607015e-34
"""Planck constant (exact) [ m² kg s⁻¹ ]"""
ħ = h/(2*π)
"""Reduced Planck constant [ m² kg s⁻¹ ]"""
k_e = 8.9875517873681764e9
"""Coulomb constant (exact) [ m³ kg s⁻² C⁻² ]"""
k_B = 1.380649e-23
"""Boltzmann constant (exact) [ m² kg K⁻¹ s⁻² ]"""
N_A = 6.02214076e23
"""Avogadro constant (exact) [ mol⁻¹ ]"""
g = 9.80665
"""Standard gravity (exact) [ m s⁻² ]"""

eV = 1.602176634e-19
"""Electronvolt (exact) [ m² kg s⁻² ]"""
q = eV
"""Elementary charge (exact) [ s A ]"""

µ_0 = 4*π*1e-7
"""Permeability of Vacuum [ kg m s⁻² A⁻² ]"""
ε_0 = 1/(µ_0*c**2)
"""Permittivity of Vacuum [ s⁴ A² kg⁻¹ m⁻³ ]"""
α = q**2/(4*π*ε_0*ħ*c)
"""Fine-structure constant [ ]"""


# astronomy
ly = 9460730472580800
"""Light-year (exact) [ m ]"""
au = 149597870700
"""Astronomical unit (exact) [ m ]"""
pc = 648000/π*au
"""Parsec (exact) [ m ]"""


# imperial units to metric (exact)
inch = 0.0254
foot = ft = 0.3048
yard = yd = 0.9144
mile = mi = 1609.344
acre = ac = 4046.8564224
pound = lb = 0.45359237
# imperial units to metric (not exact)
barrel = 0.15897
gallon_uk = 0.004546090
gallon_us = 0.003785412
ounce_int = oz = 0.028349523125
ounce_troy = toz = ozt = 0.0311034768


if __name__ == "__main__":
    print(prefix(0.9, 2))



# Copyright (C) 2016  Pachol, Vojtěch <pacholick@gmail.com>
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

from math import *                      # noqa: F401, F403
import sys as _sys
import re as _re
import numbers as _numbers

import commacalc.currencies as _currencies
import commacalc.units as _units
import commacalc.pretty_number as _pretty_number

from commacalc.currencies import *      # noqa: F401, F403
from commacalc.units import *           # noqa: F401, F403
from commacalc.extra import *           # noqa: F401, F403
# from commacalc.primes import *


_PREFIX_FUNCTIONS = {
    '√': "sqrt",
    '∛': "cbrt",
}

_POSTFIX_FUNCTIONS = {
    '²': "**2",
    '³': "**3",
}


def _replace_functions(input_):
    """replace √123 or √(123) with sqrt(123)"""
    for symbol, value in _PREFIX_FUNCTIONS.items():
        input_ = _re.sub(
            rf'{symbol}(-?[\d.]+)',
            rf'{value}(\1)',
            input_
        )
    for symbol, value in _PREFIX_FUNCTIONS.items():
        input_ = _re.sub(
            rf'{symbol}\(',
            rf'{value}(',
            input_
        )

    for symbol, value in _POSTFIX_FUNCTIONS.items():
        input_ = _re.sub(
            rf'([\d.]+){symbol}',
            rf'(\1){value}',
            input_
        )
    for symbol, value in _POSTFIX_FUNCTIONS.items():
        input_ = _re.sub(
            rf'\){symbol}',
            rf'){value}',
            input_
        )

    return input_


def main():
    input_ = ''.join(_sys.argv[1:])

    input_ = _replace_functions(input_)
    input_ = _currencies.replace_symbols(input_)
    input_ = _units.replace_symbols(input_)

    result = eval(input_)
    if isinstance(result, _numbers.Real):
        result = _pretty_number.PrettyNumber(result)
    print(result)


if __name__ == '__main__':
    _sys.exit(main())



# Copyright (C) 2016  Pachol, Vojtěch <pacholick@gmail.com>
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

import math
import itertools
from collections import Counter


class PrimesList:
    """Structure with list of primes, expanding as they're requested.
    """
    def _gen(self):
        composites = {}     # maps composites to their prime divisors
        # yield 2
        for i in itertools.count(start=3, step=2):
            prime_divisor = composites.pop(i, None)

            if prime_divisor:   # set next composite
                i += prime_divisor
                while i in composites or i & 1 == 0:
                    i += prime_divisor
                composites[i] = prime_divisor
            else:               # i is prime
                composites[i*i] = i
                self.lst.append(i)
                self.set.add(i)
                # print("g{}".format(i), end='')
                yield i

    def __init__(self):
        self.lst = [2]
        self.set = {2}
        self.gen = self._gen()

    def __iter__(self):
        """for i in primes"""
        yield from self.lst
        # yield from self.gen
        while True:
            yield next(self.gen)

    def __contains__(self, item):
        """331 in primes"""
        if self.lst[-1] >= item:
            return item in self.set

        for p in self.gen:
            if p == item:
                return True
            if p > item:
                return False
            # else continue

    def __getitem__(self, key):
        """primes[10:20]"""
        if isinstance(key, int):
            hi_idx = key
        if isinstance(key, slice):
            hi_idx = max(key.start or 0, key.stop or 0)

        while hi_idx >= len(self.lst):
            next(self.gen)
        return self.lst[key]


primes = PrimesList()


def factorize(number):
    """Factorize number to prime numbers.
    """
    primes_iter = iter(primes)
    prime = next(primes_iter)
    square_root = math.isqrt(number)

    while prime <= square_root:
        quotient, remainder = divmod(number, prime)

        if remainder:
            prime = next(primes_iter)
        else:
            yield prime
            number = quotient
            if number == 1:
                return
    yield number


_sups_table = str.maketrans("0123456789", "⁰¹²³⁴⁵⁶⁷⁸⁹")


def _to_sups(x):
    return str(x).translate(_sups_table)


def factorize_str(number):
    counter = Counter(factorize(number)).items()

    return "{} = {}".format(number, " × ".join(
        "{}{}".format(base, "" if exp == 1 else _to_sups(exp))
        for base, exp in sorted(counter)
    ))


if __name__ == "__main__":
    print(factorize_str(128))
    print(factorize_str(9))
    print(3 in primes)
    print(primes[4:])
    print(factorize_str(9699690))
    print(primes[4:])



# Copyright (C) 2016  Pachol, Vojtěch <pacholick@gmail.com>
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

import os as _os
import pickle as _pickle
from stat import ST_MTIME as _ST_MTIME
from time import time as _time
import threading as _threading


_SYMBOLS = {
    '$': "USD",
    '£': "GBP",
    '€': "EUR",
    '¥': "CNY",
    '₽': "RUB",
}


def replace_symbols(input_):
    """replace €123 with EUR*123"""
    for symbol, code in _SYMBOLS.items():
        input_ = input_.replace(symbol, code+'*')
    return input_


_RATES_URLS = [
    ("https://www.cnb.cz/cs/financni_trhy/devizovy_trh/"
     "kurzy_devizoveho_trhu/denni_kurz.txt"),
    ("https://www.cnb.cz/cs/financni_trhy/devizovy_trh/"
     "kurzy_ostatnich_men/kurzy.txt")
]
_RATES_PICKLE = _os.path.expanduser("~/.rates.pickle")

rates = {}


def _download():
    import httplib2
    import csv
    import io

    h = httplib2.Http()

    for url in _RATES_URLS:
        response, content = h.request(url)

        reader = csv.reader(io.StringIO(content.decode()), delimiter='|')
        next(reader)    # header
        next(reader)    # column names
        for row in reader:
            try:
                country, currname, amount, currcode, rate = row
                rates[currcode] = float(rate.replace(',', '.')) / int(amount)
            except ValueError:
                "Invalid row"


def _dump():
    with open(_RATES_PICKLE, mode="wb") as f:
        _pickle.dump(rates, f)


def _dl_n_dump():
    _download()
    _dump()


def _load():
    global rates

    with open(_RATES_PICKLE, mode="rb") as f:
        rates = _pickle.load(f)

    globals().update(rates)


def rate(currency):
    """Get exchange rate (ČNB)

    :param currency:    Code of currency (e.g. "USD" or "EUR")
    :returns:           Exchange rate in CZK
    """
    if not rates:
        if not _os.path.exists(_RATES_PICKLE):
            _download()
            _threading.Thread(target=_dump).start()
        elif _time() - _os.stat(_RATES_PICKLE)[_ST_MTIME] > 86400:
            _load()
            _threading.Thread(target=_dl_n_dump).start()
        else:
            _load()

    return rates[currency]


# curencies
dollar = rate("USD")
euro = rate("EUR")
sterling = rate("GBP")
ruble = rate("RUB")
yen = rate("JPY")


if __name__ == "__main__":
    print(f"USD: {USD}")    # noqa: F821
    print(f"EUR: {EUR}")    # noqa: F821
    print(f"VND: {VND}")    # noqa: F821



#!/usr/bin/python3

# Copyright (C) 2015  Pachol, Vojtěch <pacholick@gmail.com>
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

from setuptools import setup
import fastentrypoints      # noqa: F401


setup(
    name='commacalc',
    version='4.0.5',

    description="Command line calc",
    url='https://gitlab.com/pacholik1/CommaCalc',
    license='LGPL-3.0',

    author="Vojtěch Pachol",
    author_email="pacholick@gmail.com",

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3',
    ],
    keywords='python, math, calc',

    packages=['commacalc'],
    setup_requires=['fastentrypoints'],
    install_requires=['colorama', 'httplib2'],

    data_files=[('/usr/share/bash-completion/completions/',
                 ['bash_completion/,'])],

    entry_points={
        'console_scripts': [
            ',=commacalc:main',
        ],
    },
)
